/******************************************************************************
 * Implicit Euler
 * Stefan Band <bands@informatik.uni-freiburg.de>
 ******************************************************************************/
#include "ImplicitEuler.h"
#include <vector>
#include "vector.h"
#include "matrix33.h"
#include "definitions.h"


void compute_J(Mesh& m, Spring& spring)
{
	const auto& particle1 = m.vectorParticles[spring.p1];
	const auto& particle2 = m.vectorParticles[spring.p2];

	// compute jacobian for spring forces
	const auto deltaP = particle2.position - particle1.position;
	const auto distance2 = deltaP.length2();

	DclMatrix33 outerProduct = multTranspose(deltaP, deltaP);
	outerProduct /= distance2; 

	//static const DclMatrix33 id(DclVector(1, 0, 0), DclVector(0, 1, 0), DclVector(0, 0, 1));
	//spring.J = ((id - outerProduct)*(spring.initialLength / distance) - id)*(m.springConstant);

	// faster ...
	spring.J.identity();
	spring.J -= outerProduct;
	spring.J *= spring.initialLength / std::sqrt(distance2);
	spring.J(0, 0) -= DclFloat(1);
	spring.J(1, 1) -= DclFloat(1);
	spring.J(2, 2) -= DclFloat(1);
	spring.J *= m.springConstant;
}

void compute_J(Mesh& m)
{
	for (auto& spring : m.vectorSprings)
	{
		compute_J(m, spring);
	}
}


#pragma region - Matrix Implicit Euler -

typedef std::vector<DclVector> LargeVector;
typedef std::vector<DclMatrix33> LargeMatrix;


void mult_assign(LargeVector& result, const LargeMatrix& A, const LargeVector& a)
{
	const size_t n = a.size();
	for (size_t i = 0; i < n; ++i)
	{
		result[i].clear();
		for (size_t j = 0; j < n; ++j)
		{
			result[i] += mult(A[i*n + j], a[j]);
		}
	}
}

LargeVector minus_mult(const LargeVector& b, const LargeMatrix& A, const LargeVector& a)
{
	LargeVector result(a.size());
	for (size_t i = 0; i < a.size(); ++i)
	{
		for (size_t j = 0; j < a.size(); ++j)
		{
			result[i] += mult(A[i*a.size() + j], a[j]);
		}

		result[i] = b[i] - result[i];
	}
	return std::move(result);
}

void minus_eq(LargeVector& result, const LargeVector& a, const DclFloat& f)
{
	for (size_t i = 0; i < result.size(); ++i)
	{
		result[i] -= a[i]*f;
	}
}

void add_eq(LargeVector& result, const LargeVector& a, const DclFloat& f)
{
	for (size_t i = 0; i < result.size(); ++i)
	{
		result[i] += a[i]*f;
	}
}

void add_assign(LargeVector& result, const LargeVector& a, const LargeVector& b, const DclFloat& f)
{
	for (size_t i = 0; i < a.size(); ++i)
	{
		result[i] = a[i] + b[i]*f;
	}
}

DclFloat dot(const LargeVector& a, const LargeVector& b)
{
	DclFloat sum(0);
	for (size_t i = 0; i < a.size(); ++i)
	{
		sum += a[i] * b[i];
	}
	return sum;
}




void SolveConjugateGradient(const LargeMatrix& A, LargeVector& x, const LargeVector& b)
{
	auto r = std::move(minus_mult(b, A, x));
	auto d = r;
	auto delta_new = dot(r, r);

	LargeVector q(x.size());
	for (auto step = 0; step < IMPLICIT_EULER_MAX_ITERATIONS && delta_new > IMPLICIT_EULER_EPSILON; ++step)
	{
		mult_assign(q, A, d); // q = A*d

		const auto alpha = delta_new / dot(d, q);
		add_eq(x, d, alpha); // x += d*alpha
		minus_eq(r, q, alpha); // r -= q*alpha

		const auto delta = delta_new;
		delta_new = dot(r, r);

		const auto beta = delta_new / delta;
		add_assign(d, r, d, beta); // d = r + d*beta
	}
}

void integrationMatrixImplicitEuler(Mesh& m)
{
	const size_t n = m.vectorParticles.size();
	LargeMatrix A(n*n);
	LargeVector b(n);
	LargeVector x(n);

	for (size_t i = 0; i < n; ++i)
	{
		const auto& particle = m.vectorParticles[i];
		b[i] = particle.velocity*particle.mass + particle.force*m.h;
		x[i] = particle.velocity;
	}

	DclMatrix33 outerProduct;

	for (auto& spring : m.vectorSprings)
	{
		compute_J(m, spring);
		const auto J = spring.J*(m.h*m.h);

		A[spring.p1*n + spring.p1] -= J;
		A[spring.p1*n + spring.p2] += J;
		A[spring.p2*n + spring.p1] += J;
		A[spring.p2*n + spring.p2] -= J;
	}

	for (size_t i = 0; i < n; ++i)
	{
		const auto& particle = m.vectorParticles[i];

		A[i*n + i](0, 0) += particle.mass;
		A[i*n + i](1, 1) += particle.mass;
		A[i*n + i](2, 2) += particle.mass;
	}

	SolveConjugateGradient(A, x, b);

	for (size_t i = 0; i < n; ++i)
	{
		auto& particle = m.vectorParticles[i];
		particle.velocity = x[i];
		particle.position += particle.velocity*m.h;
	}
}

#pragma endregion


#pragma region - Implicit Euler -

#define Ad(particle) (particle.k1)
#define d(particle) (particle.k2)
#define r(particle) (particle.k3)
#define b(particle) (particle.k4)

void compute_Ad(Mesh& m)
{
	for (auto& particle : m.vectorParticles)
	{
		Ad(particle).clear();
	}

	DclVector j;
	for (const auto& spring : m.vectorSprings)
	{
		auto& particle1 = m.vectorParticles[spring.p1];
		auto& particle2 = m.vectorParticles[spring.p2];

		mult_assign(j, spring.J, d(particle1)); // j = J*d
		Ad(particle1) -= j;
		Ad(particle2) += j;
		
		mult_assign(j, spring.J, d(particle2)); // j = J*d
		Ad(particle1) += j;
		Ad(particle2) -= j;
	}

	for (auto& particle : m.vectorParticles)
	{
		// Ad = (M - h^2*J)d 
		//    = Md - h^2*Jd
		//    = Md + h^2(-J)d
		Ad(particle) = d(particle)*particle.mass + Ad(particle)*(m.h*m.h);
	}
}

void integrationImplicitEuler(Mesh& m)
{
	for (auto& particle : m.vectorParticles)
	{
		b(particle) = particle.velocity*particle.mass + particle.force*m.h;
		d(particle) = particle.velocity;
	}

	compute_J(m);
	compute_Ad(m);

	DclFloat delta(0);
	for (auto& particle : m.vectorParticles)
	{
		r(particle) = b(particle) - Ad(particle);
		d(particle) = r(particle);

		delta += r(particle)*r(particle);
	}

	// conjugate gradient
	for (auto step = 0; step < IMPLICIT_EULER_MAX_ITERATIONS && delta > IMPLICIT_EULER_EPSILON; ++step)
	{
		compute_Ad(m);

		DclFloat d_Ad(0);
		for (const auto& particle : m.vectorParticles)
		{
			d_Ad += d(particle)*Ad(particle);
		}

		DclFloat delta_new(0);
		const auto alpha = delta / d_Ad;
		for (auto& particle : m.vectorParticles)
		{
			particle.velocity += d(particle)*alpha;
			r(particle) -= Ad(particle)*alpha;

			delta_new += r(particle)*r(particle);
		}

		const auto beta = delta_new / delta;
		for (auto& particle : m.vectorParticles)
		{
			d(particle) = r(particle) + d(particle)*beta;
		}

		delta = delta_new;
	}

	for (auto& particle : m.vectorParticles)
	{
		particle.position += particle.velocity*m.h;
	}
}

#pragma endregion
