#pragma warning(disable:4244)
#pragma warning(disable:4018)

#include "definitions.h"
#include "textoutput.h"
#include "particle.h"
#include "tetramesh.h"

#include "precisetimer.h"
#include <algorithm>


#include "ImplicitEuler.h"

DclVector getWorld();
void loadTetraFile(Mesh& m, const char* filename);
void computeSpringSet(Mesh& m);



//////////// GLOBALS /////////////////
std::vector<Mesh> vectorMeshes;

int frame = 0;

//////////////////////////////////////

void visualizeParticles();
void computeForces(Mesh& m);

//------------------------------------------------------------------------------------------------------------------------------------

void initializeVolume(Mesh& m)
{
	for (auto& tetrahedron : m.vectorTetrahedra)
	{
		auto& particle1 = m.vectorParticles[tetrahedron.p1];
		auto& particle2 = m.vectorParticles[tetrahedron.p2];
		auto& particle3 = m.vectorParticles[tetrahedron.p3];
		auto& particle4 = m.vectorParticles[tetrahedron.p4];

		const auto e1 = particle2.position - particle1.position;
		const auto e2 = particle3.position - particle1.position;
		const auto e3 = particle4.position - particle1.position;

		tetrahedron.initialVolume = (e1 * (e2 % e3))/DclFloat(6);
	}
}

void initializeTetraMesh(Mesh& m, const DclVector& translation)
{
	for (auto& particle : m.vectorParticles)
	{
		particle.position += translation;
		particle.force.clear();
		particle.velocity.clear();
		particle.mass = 1.0;
		particle.r = m.r;
		particle.g = m.g;
		particle.b = m.b;
	}

	for (auto& spring : m.vectorSprings)
	{
		spring.initialLength = (m.vectorParticles[spring.p1].position - m.vectorParticles[spring.p2].position).length();
	}

	initializeVolume(m);
}


//------------------------------------------------------------------------------------------------------------------------------------
void initializeMesh(const char* fileName, const Integration& integration, const DclVector& translation,
					const DclFloat& h = DEFAULT_TIMESTEP,
					const DclFloat& k = DEFAULT_SPRING_CONSTANT,
					const DclFloat& ks = DEFAULT_SPRING_DAMPING_CONSTANT,
					const DclFloat& kv = DEFAULT_VOLUME_CONSTANT)
{
	vectorMeshes.push_back(Mesh());
	auto& m = vectorMeshes.back();
	m.integration = integration;
	m.h = h;
	m.springConstant = k;
	m.springDampingConstant = ks;
	m.volumeConstant = kv;
	loadTetraFile(m, fileName);
	computeSpringSet(m);
	initializeTetraMesh(m, translation);
}

// initialize the particle properties
void initializeSimulation(int argc, const char** argv)
{
	vectorMeshes.clear();

	auto param = (argc > 1) ? argv[1] : DEFAULT_MESH;

	initializeMesh(param, Integration::ImplicitEuler, DclVector(-3, 0, 0), DEFAULT_TIMESTEP, DEFAULT_SPRING_CONSTANT, 0);
	initializeMesh(param, Integration::MatrixImplicitEuler, DclVector(-1, 0, 0), DEFAULT_TIMESTEP, DEFAULT_SPRING_CONSTANT, 0);
	initializeMesh(param, Integration::Verlet, DclVector(1, 0, 0));
	initializeMesh(param, Integration::RungeKutta4, DclVector(3, 0, 0));
}

//------------------------------------------------------------------------------------------------------------------------------------

// add external forces to the sum of forces acting on the particle
void addForces(Particle& particle)
{
	particle.force += DclVector(0, -G*particle.mass, 0); // gravitational force
}

//------------------------------------------------------------------------------------------------------------------------------------

void addSpringForces(Mesh& m, const Spring& spring)
{
	auto& particle1 = m.vectorParticles[spring.p1];
	auto& particle2 = m.vectorParticles[spring.p2];

	// Compute spring force
	auto deltaP = particle1.position - particle2.position;
	const auto deltaP_length = deltaP.length();
	const auto springForce = deltaP*((DclFloat(1) - spring.initialLength/deltaP_length)*m.springConstant);

	particle1.force -= springForce;
	particle2.force += springForce;
}

void addSpringDampingForces(Mesh& m, const Spring& spring)
{
	auto& particle1 = m.vectorParticles[spring.p1];
	auto& particle2 = m.vectorParticles[spring.p2];

	auto deltaP = particle1.position - particle2.position;
	deltaP.normalize();

	const auto v1 = particle1.velocity + particle1.force*(m.h/particle1.mass);
	const auto v2 = particle2.velocity + particle2.force*(m.h/particle2.mass);

	auto deltaV = v1 - v2;
	auto f = deltaV.projectOn(deltaP);
	f *= m.springDampingConstant*deltaV.length();

	particle1.force -= f;
	particle2.force += f;
}

void addTetraForces(Mesh& m, const Tetrahedron& tetrahedron)
{
	auto& particle1 = m.vectorParticles[tetrahedron.p1];
	auto& particle2 = m.vectorParticles[tetrahedron.p2];
	auto& particle3 = m.vectorParticles[tetrahedron.p3];
	auto& particle4 = m.vectorParticles[tetrahedron.p4];

	// add forces that preserve the volume of a tetrahedron
	const auto e1 = particle2.position - particle1.position;
	const auto e2 = particle3.position - particle1.position;
	const auto e3 = particle4.position - particle1.position;

	const auto C = (e1 * (e2 % e3))/DclFloat(6) - tetrahedron.initialVolume;
	const auto f = C*m.volumeConstant;

	particle1.force += ((e2 - e1) % (e3 - e1))*f;
	particle2.force += (e3 % e2)*f;
	particle3.force += (e1 % e3)*f;
	particle4.force += (e2 % e1)*f;
}

//------------------------------------------------------------------------------------------------------------------------------------


void integrationRungeKutta4(Mesh& m)
{
	// Implement fourth-order Runge Kutta
	// use the member variables k1..k4 and l1..l4 defined in the Particle-struct
	// perform one integration step
	for (auto& particle : m.vectorParticles)
	{
		particle.positionOld = particle.position;
		particle.velocityOld = particle.velocity;

		particle.k1 = particle.velocity;
		particle.l1 = particle.force/particle.mass;

		particle.k2 = particle.k1 + particle.l1*(m.h/DclFloat(2));
		particle.velocity = particle.k2;
		particle.position = particle.positionOld + particle.k1*(m.h/DclFloat(2));
	}

	computeForces(m);

	for (auto& particle : m.vectorParticles)
	{
		particle.l2 = particle.force/particle.mass;
		particle.k3 = particle.k1 + particle.l2*(m.h/DclFloat(2));
		particle.velocity = particle.k3;
		particle.position = particle.positionOld + particle.k2*(m.h/DclFloat(2));
	}

	computeForces(m);

	for (auto& particle : m.vectorParticles)
	{
		particle.l3 = particle.force/particle.mass;
		particle.k4 = particle.k1 + particle.l3*m.h;
		particle.velocity = particle.k4;
		particle.position = particle.positionOld + particle.k3*m.h;
	}

	computeForces(m);

	for (auto& particle : m.vectorParticles)
	{
		particle.l4 = particle.force/particle.mass;

		particle.position = particle.positionOld + (particle.k1 + (particle.k2 + particle.k3)*DclFloat(2) + particle.k4)*(m.h/DclFloat(6));
		particle.velocity = particle.velocityOld + (particle.l1 + (particle.l2 + particle.l3)*DclFloat(2) + particle.l4)*(m.h/DclFloat(6));
	}
}

// perform one time-integration step
void integrate(const Mesh& m, Particle& particle)
{
	switch (m.integration)
	{
	case Integration::Euler:
		particle.position += particle.velocity*m.h;
		particle.velocity += particle.force*(m.h/particle.mass);
		break;

	case Integration::EulerCromer:
		particle.velocity += particle.force*(m.h/particle.mass);
		particle.position += particle.velocity*m.h;
		break;

	case Integration::Verlet:
		{
			auto newPosition = particle.position + particle.velocity*m.h + particle.force*(m.h*m.h/particle.mass);
			particle.velocity = (newPosition - particle.position)/m.h;
			particle.position = newPosition;
		}
		break;
	}
}

//------------------------------------------------------------------------------------------------------------------------------------

// collision handling with the world
void worldCollision(Particle& particle)
{
	auto world = getWorld();
	world *= 0.5;

#define HANDLECOLLISION(axis) \
	if (particle.position.axis < -world.axis) \
	{ \
		particle.velocity.axis *= -WORLD_ELASTICITY; \
		particle.position.axis = -world.axis; \
	} \
	else if (particle.position.axis > world.axis) \
	{ \
		particle.velocity.axis *= -WORLD_ELASTICITY; \
		particle.position.axis = world.axis; \
	}

	HANDLECOLLISION(x);
	HANDLECOLLISION(y);
	HANDLECOLLISION(z);
}

//------------------------------------------------------------------------------------------------------------------------------------

// compute and output the total energies in the scene
void computeEnergies(const Mesh& m, int i)
{
	DclFloat e_kin = 0, e_pot = 0, e_def = 0;
	auto world = getWorld();

	for (const auto& particle : m.vectorParticles)
	{
		e_kin += particle.velocity.length2()/DclFloat(2)*particle.mass;
		e_pot += (particle.position.y + world.z/DclFloat(2))*particle.mass*G;
	}

	// Compute deformation energy
	for (const auto& spring : m.vectorSprings)
	{
		const auto dx = (m.vectorParticles[spring.p1].position - m.vectorParticles[spring.p2].position).length() - spring.initialLength;
		e_def += m.springConstant*(dx*dx)/DclFloat(2);
	}

	Textout::printvar(6 + i, 0, "Ekin %f, Epot %f, Edef %f, Sum %f", e_kin, e_pot, e_def, e_kin + e_pot + e_def);
}

//------------------------------------------------------------------------------------------------------------------------------------

void computeForces(Mesh& m)
{
	// reset the vector of external forces
	for (auto& particle : m.vectorParticles)
	{
		particle.force.clear();
	}

	// compute external forces on the particles
	for (auto& particle : m.vectorParticles)
	{
		addForces(particle);
	}

	for (const auto& spring : m.vectorSprings)
	{
		addSpringForces(m, spring);
	}
	
	for (const auto& tetra : m.vectorTetrahedra)
	{
		addTetraForces(m, tetra);
	}

	for (const auto& spring : m.vectorSprings)
	{
		addSpringDampingForces(m, spring);
	}
}

// perform one simulation step
void simulationStep()
{
	#pragma omp parallel for
	for (int mi = 0; mi < vectorMeshes.size(); ++mi)
	{
		auto& m = vectorMeshes[mi];

		computeForces(m);

		// compute collision impulses
		for (auto& particle : m.vectorParticles)
		{
			worldCollision(particle);
		}

		// perform one integration step
		switch (m.integration)
		{
		case Integration::RungeKutta4:
			integrationRungeKutta4(m);
			break;

		case Integration::ImplicitEuler:
			integrationImplicitEuler(m);
			break;

		case Integration::MatrixImplicitEuler:
			integrationMatrixImplicitEuler(m);
			break;

		default:
			for (auto& particle : m.vectorParticles)
			{
				integrate(m, particle);
			}
		}
	}

#ifdef FIXED_FRAME_RATE
	static performance_counter::value_type timeLastRender(0);
	static auto renderDuration = performance_counter::from_ms(DESIRED_FRAME_DURATION);
	static size_t frameCounter = 0;

	++frameCounter;

	auto now = performance_counter::value();
	auto renderDiff = now - timeLastRender;
	if (renderDiff >= renderDuration)
	{
#endif
		int index = 0;
		for (auto& m : vectorMeshes)
		{
			// compute and output energies
			computeEnergies(m, index++);
		}

		// draw particles
		visualizeParticles();

#ifdef FIXED_FRAME_RATE
		Textout::printvar(5, 0, "Frame %d, Frames/second %.2f, Num updates before render %d", frame++, 1000.0/performance_counter::ms(renderDiff), frameCounter);
		frameCounter = 0;
		timeLastRender = performance_counter::value();
	}
#endif
}


